-- This module provides basic definitions in propositional logic.

-- Agda is based on the propositions-as-types paradigm, also known as
-- the Curry-Howard isomorphism. Consequently, each proposition is
-- identified with the set of its proofs.

module Lib.Proposition where

-- ----------------------------------------------------------------------
-- * Definitions

infixr 2 _∧_ _×_ _∪_
infixr 4 _,_

-- The proposition "True" has a canonical proof, so it is identified
-- with the singleton type.
record True : Set where
  constructor <>
    -- No fields

-- The proposition "False" has no proof, so it is identified with the
-- empty type.
data False : Set where
  -- No constructors.

-- Existential quantifier. A proof of "there exists x : A such that
-- B(x)" is a pair whose first component is x and whose second
-- component is a proof of B(x).
data Exists {l} (A : Set l) (B : A -> Set) : Set l where
  _,_ : (x : A) -> B x -> Exists A B

open Exists public

-- Get the first component of a pair.
fst : ∀ {A : Set} {B : A -> Set} -> Exists A B -> A
fst (x , y) = x

-- Get the second component of a pair.
snd : ∀ {A : Set} {B : A -> Set} -> (x : Exists A B) -> B (fst x)
snd (x , y) = y

-- Conjunction is a special case.
_∧_ : ∀ (A : Set) (B : Set) -> Set
A ∧ B = Exists A (λ x -> B)

-- Negation.
Not : Set -> Set
Not A = A -> False

-- ----------------------------------------------------------------------
-- * Alternate names

-- Cartesian product.
_×_ : ∀ (A : Set) (B : Set) -> Set
_×_ = _∧_

-- Nicer syntax for "exists".
∃ : ∀ {l} {A : Set l} (B : A -> Set) -> Set l
∃ = Exists _

-- ----------------------------------------------------------------------
-- * Sum

infix 9 ⌜_
infix 10 _⌝

data _∪_ {l} (A B : Set l) : Set l where
  ⌜_ : A -> A ∪ B
  _⌝ : B -> A ∪ B
  
