-- This module provides two versions of the Reidemeister-Schreier
-- theorem for monoids, and shows that the Reidemeister-Schreier
-- procedure can be composed.
--
-- * The simplified Reidemeister-Schreier theorem assumes that X and Y
--   are sets of generators for the same monoid. In this case, there
--   are translations from X to Y* and from Y to X*, and no cosets are
--   required.
--
-- * The full Reidemeister-Schreier theorem is Theorem 4.2 from
--   arXiv:2204.02217. It assumes that Y generates some monoid and X
--   generates a submonoid, as described in the paper.

open import Lib.Proposition
open import Lib.Equality as Eq using (_==_)
open import Lib.Inspect

open import Lib.Word
open import Lib.Word-Lemmas

module Lib.Reidemeister-Schreier where

-- ----------------------------------------------------------------------
-- * The simplified Reidemeister-Schreier theorem for monoids

-- We assume that X and Y are sets of generators, Γ and Δ are sets of
-- relations over X and Y, respectively, and there are translations
-- f : X -> Y* and g : Y -> X* compatible with Γ and Δ. In this case,
-- it follows that ⟨X | Γ⟩ and ⟨Y | Δ⟩ are isomorphic monoids.

module Reidemeister-Schreier-Simplified
       {X : Set} {Γ : Context X} 
       {Y : Set} {Δ : Context Y} 
       (f : X -> Word Y)
       (g : Y -> Word X)
       (hypA : ∀ (x : X) -> Γ ⊢ gen x === (g *) (f x))
       (hypB : ∀ (u t : Word Y) -> u === t ∈ Δ -> Γ ⊢ (g *) u === (g *) t)
  where

  open Monoid-Equational

  lemma-a : ∀ (w : Word X) -> Γ ⊢ w === (g *) ((f *) w)
  lemma-a (gen x) = hypA x
  lemma-a ε = refl
  lemma-a (w • v) = cong (lemma-a w) (lemma-a v)

  lemma-b : ∀ {u t : Word Y} -> Δ ⊢ u === t -> Γ ⊢ (g *) u === (g *) t
  lemma-b (axiom x) = hypB _ _ x
  lemma-b refl = refl
  lemma-b (symm hyp) = lemma-b hyp reversed
  lemma-b (trans hyp1 hyp2) = trans (lemma-b hyp1) (lemma-b hyp2)
  lemma-b (cong hyp1 hyp2) = cong (lemma-b hyp1) (lemma-b hyp2)
  lemma-b assoc = assoc
  lemma-b left-unit = left-unit
  lemma-b right-unit = right-unit

  reidemeister-schreier-simplified : (w v : Word X) -> Δ ⊢ (f *) w === (f *) v -> Γ ⊢ w === v
  reidemeister-schreier-simplified w v hyp = 
    equational w
            by lemma-a w
        equals (g *) ((f *) w)
            by lemma-b hyp
        equals (g *) ((f *) v)
            by lemma-a v reversed
        equals v

-- ----------------------------------------------------------------------
-- * The full Reidemeister-Schreier theorem for monoids

-- We start with some notations.

-- Extend a function from generators to words, with accumulator.
_** : {Y X C : Set} -> (C -> Y -> Word X × C) -> (C -> Word Y -> Word X × C)
_** h c (gen y) = h c y
_** h c ε = (ε , c)
_** h c (w • u) with (h **) c w
_** h c (w • u) | (w' , c') with (h **) c' u
_** h c (w • u) | (w' , c') | (u' , c'') = (w' • u' , c'')

-- Abbreviation: We write Γ ⊢ₚ (u , c) === (t , c')
-- for Γ ⊢ u === t and c == c'.
_⊢ₚ_===_ : {X C : Set} (Γ : Context X) -> Word X × C -> Word X × C -> Set
Γ ⊢ₚ (u , c) === (t , c') = Γ ⊢ u === t ∧ c == c'

-- Reflexivity, symmetry, and transitivity for ⊢ₚ.
lemma-refl-⊢ₚ : ∀ {X Γ C} {p q : Word X × C} -> p == q -> Γ ⊢ₚ p === q
lemma-refl-⊢ₚ {p = u1 , c1} {u2 , c2} Eq.refl = (refl , Eq.refl)

lemma-symm-⊢ₚ : ∀ {X Γ C} {p q : Word X × C} -> Γ ⊢ₚ p === q -> Γ ⊢ₚ q === p
lemma-symm-⊢ₚ {p = u1 , c1} {u2 , c2} (hyp1 , eq1) = (symm hyp1 , Eq.symm eq1)

lemma-trans-⊢ₚ : ∀ {X Γ C} {p q r : Word X × C} -> Γ ⊢ₚ p === q -> Γ ⊢ₚ q === r -> Γ ⊢ₚ p === r
lemma-trans-⊢ₚ {p = u1 , c1} {u2 , c2} {u3 , c3} (hyp1 , eq1) (hyp2 , eq2) = (trans hyp1 hyp2 , Eq.trans eq1 eq2)

-- Two obvious lemmas about ⊢ₚ.
lemma-⊢ₚ-fst : ∀ {X Γ C} (lhs rhs : Word X × C) -> Γ ⊢ₚ lhs === rhs -> Γ ⊢ fst lhs === fst rhs
lemma-⊢ₚ-fst (w , c) (w' , c') hyp = fst hyp

lemma-⊢ₚ-snd : ∀ {X Γ C} (lhs rhs : Word X × C) -> Γ ⊢ₚ lhs === rhs -> snd lhs == snd rhs
lemma-⊢ₚ-snd (w , c) (w' , c') hyp = snd hyp

-- The Reidemeister-Schreier theorem. We assume that X and Y are sets
-- of generators, Γ and Δ are sets of relations over X and Y,
-- respectively, C is a set (typically a set of coset representatives,
-- although this is not assumed), and there are translations
-- f : X -> Y* and h : C × Y -> X* × C, subject to two assumptions.
-- In that case, Γ is a complete set of relations for the submonoid of
-- Y* generated by X.
module Reidemeister-Schreier-Full
       {X : Set} {Γ : Context X}
       {Y : Set} {Δ : Context Y}
       (C : Set)
       (I : C)
       (f : X -> Word Y)
       (h : C -> Y -> Word X × C)
       (hypA : ∀ (x : X) -> Γ ⊢ₚ (h **) I (f x) === (gen x , I))
       (hypB : ∀ (c : C) {w w' : Word Y} -> w === w' ∈ Δ -> Γ ⊢ₚ (h **) c w === (h **) c w')
  where

    open Monoid-Equational

    -- Definition: A word is special if it doesn't leave the "I"-coset.
    special : Word Y -> Set
    special w = snd ((h **) I w) == I

    -- Lemma: ε is special.
    lemma-special-ε : special ε
    lemma-special-ε = Eq.refl

    -- Lemma: Special words are closed under multiplication.
    lemma-special-• : ∀ (w v : Word Y) -> special w -> special v -> special (w • v)
    lemma-special-• w v hyp-w hyp-v with (h **) I w | inspect ((h **) I) w
    lemma-special-• w v hyp-w hyp-v | (w' , c') | [ eq1 ] with (h **) c' v | inspect ((h **) c') v
    lemma-special-• w v hyp-w hyp-v | (w' , c') | [ eq1 ] | (v' , c'') | [ eq2 ] = lem3
      where
          lem1 : c' == I
          lem1 = hyp-w

          lem2 : (v' , c'') == (h **) I v
          lem2 =
            Eq.equational (v' , c'')
                    Eq.by Eq.symm eq2
                   equals (h **) c' v
                    Eq.by Eq.context (λ □ → (h **) □ v) lem1
                   equals (h **) I v

          lem3 : c'' == I
          lem3 = 
            Eq.equational c''
                    Eq.by Eq.context snd lem2
                   equals snd ((h **) I v)
                    Eq.by hyp-v
                   equals I

    -- Lemma: The image of f is special.
    lemma-special-f : ∀ (x : X) -> special (f x)
    lemma-special-f x = lemma-⊢ₚ-snd ((h **) I (f x)) (gen x , I) (hypA x)

    -- Lemma: The image of (f *) is special.
    lemma-special-f* : ∀ (w : Word X) -> special ((f *) w)
    lemma-special-f* (gen x) = lemma-special-f x
    lemma-special-f* ε = lemma-special-ε
    lemma-special-f* (w • v) = lemma-special-• ((f *) w) ((f *) v) (lemma-special-f* w) (lemma-special-f* v)

    -- Definition: For special words, there is a translation back.
    g : Word Y -> Word X
    g w = fst ((h **) I w)

    -- Lemma: g preserves ε.
    lemma-g-ε : g ε == ε
    lemma-g-ε = Eq.refl

    -- Lemma: g is a homomorphism on special words.
    lemma-g-• : ∀ (w v : Word Y) -> special w -> g (w • v) == g w • g v
    lemma-g-• w v hyp with (h **) I w | inspect ((h **) I) w
    lemma-g-• w v hyp | (w' , c') | [ eq1 ] with (h **) c' v | inspect ((h **) c') v
    lemma-g-• w v hyp | (w' , c') | [ eq1 ] | (v' , c'') | [ eq2 ] = lem6
      where
          lem1 : c' == I
          lem1 = hyp

          lem2 : (v' , c'') == (h **) I v
          lem2 =
            Eq.equational (v' , c'')
                    Eq.by eq2 Eq.reversed
                   equals (h **) c' v
                    Eq.by Eq.context (λ □ → (h **) □ v) lem1
                   equals (h **) I v

          lem4 : v' == g v
          lem4 = Eq.context fst lem2 

          lem6 : w' • v' == w' • g v
          lem6 = Eq.context (λ □ → w' • □) lem4

    -- Lemma: g is a left inverse of f *.
    lemma-a : ∀ (w : Word X) -> Γ ⊢ w === g ((f *) w)
    lemma-a (gen x) = lemma-⊢ₚ-fst (gen x , I) ((h **) I (f x)) (lemma-symm-⊢ₚ (hypA x))
    lemma-a ε = refl
    lemma-a (w • u) = 
      equational w • u
              by left lemma-a w
          equals g ((f *) w) • u
              by right lemma-a u
          equals g ((f *) w) • g ((f *) u)
              by refl' (lemma-g-• ((f *) w) ((f *) u) (lemma-special-f* w)) reversed
          equals g ((f *) w • (f *) u)

    -- Lemma: Hypothesis B can be extended from the elements of Δ to
    -- all consequences of Δ.
    lemma-hypB : ∀ (c : C) (u t : Word Y) -> Δ ⊢ u === t -> Γ ⊢ₚ (h **) c u === (h **) c t

    lemma-hypB c u t (axiom x) = hypB c x

    lemma-hypB c u .u refl = lemma-refl-⊢ₚ Eq.refl

    lemma-hypB c ((u • v) • w) (.u • (.v • .w)) assoc with (h **) c u
    ... | (u' , c') with (h **) c' v
    ... | (v' , c'') with (h **) c'' w
    ... | (w' , c''') = (assoc , Eq.refl)

    lemma-hypB c (ε • u) .u left-unit with (h **) c u
    ... | (u' , c') = (left-unit , Eq.refl)

    lemma-hypB c (u • ε) .u right-unit with (h **) c u
    ... | (u' , c') = (right-unit , Eq.refl)

    lemma-hypB c u t (symm hyp) = lemma-symm-⊢ₚ (lemma-hypB c t u hyp)

    lemma-hypB c u t (trans {v = v} hyp1 hyp2) = lemma-trans-⊢ₚ (lemma-hypB c u v hyp1) (lemma-hypB c v t hyp2)

    lemma-hypB c (u • u') (t • t') (cong hyp1 hyp2)
      with (h **) c u | (h **) c t | lemma-hypB c u t hyp1
    ... | (u'' , c') | (t'' , c'') | (ih1 , ih1')
      rewrite ih1'
      with (h **) c'' u' | (h **) c'' t' | lemma-hypB c'' u' t' hyp2
    ... | (u''' , c''') | (t''' , c'''') | (ih2 , ih2') = (cong ih1 ih2 , ih2')

    -- Lemma: g preserves relations.
    lemma-b : ∀ (u t : Word Y) -> Δ ⊢ u === t -> Γ ⊢ g u === g t
    lemma-b u t hyp with (h **) I u | (h **) I t | lemma-hypB I u t hyp
    ... | (u' , c') | (t' , c'') | (ih , ih') = ih

    -- Reidemeister-Schreier Theorem.
    reidemeister-schreier : (w v : Word X) -> Δ ⊢ (f *) w === (f *) v -> Γ ⊢ w === v
    reidemeister-schreier w v hyp =
      equational w
              by lemma-a w
          equals g ((f *) w)
              by lemma-b ((f *) w) ((f *) v) hyp
          equals g ((f *) v)
              by lemma-a v reversed
          equals v

-- ----------------------------------------------------------------------
-- * The composition of multiple applications of the
--   Reidemeister-Schreier theorem

-- This lemma states that when we compose two translations from
-- generators to words, the composition behaves as expected, i.e.,
--
-- g* ∘ f* = (g* ∘ f)*.
lemma-*-* : ∀ {X Y Z : Set} -> (f : X -> Word Y) -> (g : Y -> Word Z) -> ∀ (w : Word X) -> (g *) ((f *) w) == ((\y -> (g *) (f y)) *) w
lemma-*-* f g (gen x) = Eq.refl
lemma-*-* f g ε = Eq.refl
lemma-*-* f g (w • w') = Eq.cong _•_ (lemma-*-* f g w) (lemma-*-* f g w')


-- Two different way of composing two functions.
infix 6 _◐_ _◑_  -- Type these as: \ci7 \ci8
_◐_ : ∀ {X Y Z C : Set} -> (g : Word Y -> Word Z) -> (f : C -> X -> Word Y × C) -> (C -> X -> Word Z × C)
_◐_ g f c x = g (fst (f c x)) , snd (f c x)

_◑_ : ∀ {X Y Z C : Set} -> (g : C -> Word Y -> Word Z × C) -> (f : X -> Word Y) -> (C -> X -> Word Z × C)
_◑_ g f c x = g c (f x)

-- Properties enjoyed by the composition.
-- g * ◐ f ** = (g * ◐ f) **.
lemma-*-** : ∀ {X Y Z C : Set} -> (g : Y -> Word Z) -> (f : C -> X -> Word Y × C) -> ∀ (c : C) (w : Word X) ->  ((g *) ◐ (f **) ) c w == (((g *) ◐ f) **) c w
lemma-*-** g f c (gen x) = Eq.refl
lemma-*-** g f c ε = Eq.refl
lemma-*-** g f c (w • v) with (f **) c w | inspect ((f **) c) w
... | w' , c' | [ eqw ] with (f **) c' v | inspect ((f **) c') v
... | v' , c'' | [ eqv ] with lemma-*-** g f c w | lemma-*-** g f c' v 
... | ih1 | ih2 rewrite eqw | eqv | Eq.symm ih1 | Eq.symm ih2 = Eq.refl

-- g ** ◑ f* = (g ** ◑ f) **.
lemma-**-* : ∀ {X Y Z C : Set} -> (g : C -> Y -> Word Z × C) -> (f : X -> Word Y) -> ∀ (c : C) (w : Word X) ->  ((g **) ◑ (f *)) c w == (((g **) ◑ f) **) c w
lemma-**-* g f c (gen x) = Eq.refl
lemma-**-* g f c ε = Eq.refl
lemma-**-* g f c (w • v) with (g **) c ((f *) w) | inspect ((g **) c) ((f *) w)
... | w' , c' | [ eqw ] with (g **) c' ((f *) v) | inspect ((g **) c') ((f *) v)
... | v' , c'' | [ eqv ] with lemma-**-* g f c w | lemma-**-* g f c' v 
... | ih1 | ih2 rewrite eqw | eqv | Eq.symm ih1 | Eq.symm ih2 = Eq.refl

-- Composing two applications of the simplified Reidemeister-Schreier
-- procedure. Note that the roles of X (Γ) and Y (Δ) are switched
-- compared to the module Reidemeister-Schreier-Simplified.

module Reidemeister-Schreier-SS
       {X : Set} {Γ : Context X} 
       {Y : Set} {Δ : Context Y} 
       {Z : Set} {ζ : Context Z} 
       (f : X -> Word Y)
       (g : Y -> Word X)
       (h : Y -> Word Z)
       (k : Z -> Word Y)
       (hypA1 : ∀ (y : Y) -> Δ ⊢ gen y === (f *) (g y))
       (hypB1 : ∀ (u t : Word X) -> u === t ∈ Γ -> Δ ⊢ (f *) u === (f *) t)
       (hypA2 : ∀ (z : Z) -> ζ ⊢ gen z === (h *) (k z))
       (hypB2 : ∀ (u t : Word Y) -> u === t ∈ Δ -> ζ ⊢ (h *) u === (h *) t)
  where

  module R1 = Reidemeister-Schreier-Simplified {Y} {Δ} {X} {Γ} g f hypA1 hypB1
  module R2 = Reidemeister-Schreier-Simplified {Z} {ζ} {Y} {Δ} k h hypA2 hypB2
  
  hf : X -> Word Z
  hf x = (h *) (f x)

  gk : Z -> Word X
  gk z = (g *) (k z)

  hypA3 : ∀ (z : Z) -> ζ ⊢ gen z === (hf *) (gk z)
  hypA3 z with R1.lemma-a (k z)
  ... | hyp with hypA2 z
  ... | hyp2 with lemma-*-* f h ((g *) (k z))
  ... | hyp3 rewrite (Eq.symm hyp3) = trans hyp2 (R2.lemma-b hyp)

  hypB3 : ∀ (u t : Word X) -> u === t ∈ Γ -> ζ ⊢ (hf *) u === (hf *) t
  hypB3 u t ax with hypB1 u t ax
  ... | h1 with R2.lemma-b h1
  ... | h2 with lemma-*-* f h u | lemma-*-* f h t
  ... | h3 | h4 rewrite (Eq.symm h3) | (Eq.symm h4) = h2

  open Reidemeister-Schreier-Simplified {Z} {ζ} {X} {Γ} gk hf hypA3 hypB3 public

-- Composing the simplified and full Reidemeister-Schreier procedures
-- (full followed by simplified). Note that the roles of X (Γ) and Y
-- (Δ) are switched compared to the module Reidemeister-Schreier-Full.

module Reidemeister-Schreier-SF
       {X : Set} {Γ : Context X} 
       {Y : Set} {Δ : Context Y} 
       {Z : Set} {ζ : Context Z}
       (C : Set)
       (I : C)
       (f : C -> X -> Word Y × C)
       (g : Y -> Word X)
       (h : Y -> Word Z)
       (k : Z -> Word Y)
       (hypA1 : ∀ (y : Y) -> Δ ⊢ₚ (f **) I (g y) === (gen y , I))
       (hypB1 : ∀ (c : C) {w w' : Word X} -> w === w' ∈ Γ -> Δ ⊢ₚ (f **) c w === (f **) c w')
       (hypA2 : ∀ (z : Z) -> ζ ⊢ gen z === (h *) (k z))
       (hypB2 : ∀ (u t : Word Y) -> u === t ∈ Δ -> ζ ⊢ (h *) u === (h *) t)
  where

  module R1 = Reidemeister-Schreier-Full {Y} {Δ} {X} {Γ} C I g f hypA1 hypB1
  module R2 = Reidemeister-Schreier-Simplified {Z} {ζ} {Y} {Δ} k h hypA2 hypB2
  
  hf : C -> X -> Word Z × C
  hf c x = ((h *) (fst (f c x))) , (snd (f c x))

  gk : Z -> Word X
  gk z = (g *) (k z)

  hypA3 : ∀ (z : Z) -> ζ ⊢ₚ (hf **) I (gk z) === (gen z , I)
  hypA3 z with R1.lemma-a (k z)
  ... | hyp with hypA2 z
  ... | hyp2 with lemma-*-** h f I ((g *) (k z))
  ... | hs rewrite (Eq.symm hs) with R1.lemma-special-f* (k z)
  ... | sfs' = trans (R2.lemma-b (symm hyp)) (symm hyp2) , sfs'
  
  hypB3 : ∀ (c : C) {w w' : Word X} -> w === w' ∈ Γ -> ζ ⊢ₚ (hf **) c w === (hf **) c w'
  hypB3 c {u} {t} ax with hypB1 c ax
  ... | h1 with lemma-⊢ₚ-fst ((f **) c u) ((f **) c t) h1 | lemma-⊢ₚ-snd ((f **) c u) ((f **) c t) h1
  ... | fh | sh with R2.lemma-b fh | lemma-*-** h f c u | lemma-*-** h f c t
  ... | h2 | ss1 | ss2 rewrite (Eq.symm ss1) | (Eq.symm ss2) = R2.lemma-b fh , sh

  open Reidemeister-Schreier-Full {Z} {ζ} {X} {Γ} C I gk hf hypA3 hypB3 public


-- Composing the simplified and full Reidemeister-Schreier procedures
-- (simplified followed by full). Note that the roles of X (Γ) and Y
-- (Δ) are switched compared to the module Reidemeister-Schreier-Simplified.

module Reidemeister-Schreier-FS
       {X : Set} {Γ : Context X} 
       {Y : Set} {Δ : Context Y} 
       {Z : Set} {ζ : Context Z}
       (C : Set)
       (I : C)
       (f : X -> Word Y)
       (g : Y -> Word X)
       (h : C -> Y -> Word Z × C)
       (k : Z -> Word Y)
       (hypA1 : ∀ (y : Y) -> Δ ⊢ gen y === (f *) (g y))
       (hypB1 : ∀ (u t : Word X) -> u === t ∈ Γ -> Δ ⊢ (f *) u === (f *) t)
       (hypA2 : ∀ (z : Z) -> ζ ⊢ₚ (h **) I (k z) === (gen z , I))
       (hypB2 : ∀ (c : C) {w w' : Word Y} -> w === w' ∈ Δ -> ζ ⊢ₚ (h **) c w === (h **) c w')
  where

  module R1 = Reidemeister-Schreier-Simplified {Y} {Δ} {X} {Γ} g f hypA1 hypB1
  module R2 = Reidemeister-Schreier-Full {Z} {ζ} {Y} {Δ} C I k h hypA2 hypB2
  
  hf : C -> X -> Word Z × C
  hf = (h **) ◑ f

  gk : Z -> Word X
  gk z = (g *) (k z)

  hypA3 : ∀ (z : Z) -> ζ ⊢ₚ (hf **) I (gk z) === (gen z , I)
  hypA3 z with R1.lemma-a (k z)
  ... | hyp with hypA2 z
  ... | hyp2 with lemma-**-* h f I ((g *) (k z))
  ... | hs rewrite (Eq.symm hs) | R2.lemma-special-f z with R2.lemma-hypB I (k z) ((f *) ((g *) (k z))) hyp
  ... | hyp' = lemma-trans-⊢ₚ (lemma-symm-⊢ₚ hyp') hyp2

  hypB3 : ∀ (c : C) {w w' : Word X} -> w === w' ∈ Γ -> ζ ⊢ₚ (hf **) c w === (hf **) c w'
  hypB3 c {u} {t} ax with hypB1 u t ax
  ... | h1 with R2.lemma-hypB c ((f *) u) ((f *) t) h1 | lemma-**-* h f c u | lemma-**-* h f c t
  ... | h2 | ss1 | ss2 rewrite (Eq.symm ss1) | (Eq.symm ss2) = h2

  open Reidemeister-Schreier-Full {Z} {ζ} {X} {Γ} C I gk hf hypA3 hypB3 public

